package main

import (
	"log"
	"net/http"
	"os"
	"query-service/internal/drivers/rpcdriver"
	"query-service/internal/usecases/consume"
	"query-service/internal/usecases/databaseinfo"
	"query-service/internal/usecases/produce"
	"query-service/pkg/logger"

	"git.science.uu.nl/datastrophe/broker"
	"git.science.uu.nl/datastrophe/keyvaluestore"
	"git.science.uu.nl/datastrophe/objectstore"
	"git.science.uu.nl/datastrophe/query-conversion"
	"git.science.uu.nl/datastrophe/query-conversion/aql"
	"git.science.uu.nl/datastrophe/query-conversion/cypher"
	queryexecution "git.science.uu.nl/datastrophe/query-execution"
	"git.science.uu.nl/datastrophe/query-execution/arangodb"
	"git.science.uu.nl/datastrophe/query-execution/neo4j"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"github.com/thijsheijden/alice"
)

func main() {
	alice.SetLogging()
	logger.Start()

	// MARK: Create relevant services
	keyValueStore := keyvaluestore.NewDriver()

	objectStore := objectstore.NewDriver()
	err := objectStore.Connect()
	if err != nil {
		logger.Log(err.Error())
	}

	// Create new rpc driver
	rpcDriver := rpcdriver.New()

	// MARK: Create alice RabbitMQ services
	broker := broker.NewDriver()

	// Instantiate an implementation of the produce UseCase
	produceService := produce.NewService(broker, keyValueStore)

	var queryConverter query.Converter
	// Choose a query conversion service based on environment variables
	switch os.Getenv("QUERY_CONVERSION_LANGUAGE") {
	case "aql":
		queryConverter = aql.NewService()
	case "cypher":
		queryConverter = cypher.NewService()
	}

	var queryExecutor queryexecution.Executor
	// Choose a query executor based on environment variables
	switch os.Getenv("QUERY_EXECUTION") {
	case "arangodb":
		queryExecutor = arangodb.NewService()
	case "neo4j":
		queryExecutor = neo4j.NewService()
	}

	databaseInfoService := databaseinfo.NewService(rpcDriver)

	consumeService := consume.NewService(objectStore, broker, produceService, queryConverter, queryExecutor, databaseInfoService)

	// MARK: Start services
	keyValueStore.Connect()
	produceService.Start()

	go consumeService.Start()

	// Expose /metrics endpoint for prometheus
	logger.Log("Starting metrics server")
	http.Handle("/metrics", promhttp.Handler())
	log.Fatal(http.ListenAndServe(":8080", nil))

	select {}
}
