package consume

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"query-service/internal/entity"
	"query-service/pkg/errorhandler"
	"strings"
	"time"

	"git.science.uu.nl/datastrophe/broker"
	"git.science.uu.nl/datastrophe/query-conversion"
)

// HandleMessage gets called when a message is received
func (s *Service) HandleMessage(msg *broker.Message) {
	// Grab sessionID and clientID from the headers
	sessionID, ok := msg.Headers["sessionID"].(string)
	if !ok {
		return
	}

	clientID, ok := msg.Headers["clientID"].(string)
	if !ok {
		return
	}

	var queryID string
	queryID, ok = msg.Headers["queryID"].(string)
	if !ok {
		queryID = ""
	}

	s.sendStatus("Received", &sessionID, &queryID)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	JSONQuery, err := query.UnmarshalJSON(&msg.Body)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_translation_error",
			Value: "failed to unmarshall the incoming JSON query",
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	// Check if there was a database name
	if JSONQuery.DatabaseName == "" {
		errorMsg := entity.MessageStruct{
			Type:  "query_malformed_request_error",
			Value: "no database name supplied",
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	s.sendStatus("Translating", &sessionID, &queryID)

	// Convert the json byte msg to a query string
	query, err := s.queryConverter.ConvertQuery(JSONQuery)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_translation_error",
			Value: err.Error(),
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	// TODO: Figure out a way to do this in the integration tests instead
	// Remove newlines and tabs from string
	queryNoNewlines := strings.ReplaceAll(*query, "\n", "")
	queryNoTabs := strings.ReplaceAll(queryNoNewlines, "\t", "")

	// Send the resulting query back to the client
	translationMsg := entity.MessageStruct{
		Type:  "query_translation_result",
		Value: queryNoTabs,
	}
	translationMsgBytes, err := json.Marshal(translationMsg)
	if err != nil {
		errorhandler.LogError(err, "Marshalling query_translation_result went wrong!") // TODO: send error message to client instead
		return
	}
	s.producer.PublishMessage(&translationMsgBytes, &sessionID) // TODO: should this be a go routine?

	s.sendStatus("Getting database credentials", &sessionID, &queryID)

	// Fetch database credentials from the user service
	databaseInfo, err := s.databaseInfoService.GetDatabaseInfo(&clientID, &JSONQuery.DatabaseName)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_database_not_found",
			Value: err.Error(),
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	s.sendStatus("Processing on database", &sessionID, &queryID)

	// execute and retrieve result
	// convert result to general (node-link) format
	result, err := s.requestSender.ExecuteQuery(*query, databaseInfo.Username, databaseInfo.Password, databaseInfo.URL, databaseInfo.Port, databaseInfo.InternalDatabaseName)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_database_error",
			Value: err.Error(),
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	s.sendStatus("Completed", &sessionID, &queryID)

	// Add type indicator to result from database
	var res interface{}
	json.Unmarshal(*result, &res)
	resultMsg := entity.MessageStruct{
		Type:  "query_result",
		Value: res,
	}
	resultMsgBytes, err := json.Marshal(resultMsg)
	if err != nil {
		errorhandler.LogError(err, "Marshalling query_result went wrong!") // TODO: send error message to client instead
		return
	}

	// Publish the result
	s.producer.PublishMessage(&resultMsgBytes, &sessionID)

	// Caching message in object store
	if queryID != "" {
		cacheQueryResultContext, cancelCacheQueryResult := context.WithTimeout(context.Background(), time.Second*20)
		defer cancelCacheQueryResult()
		s.objectStore.Put(cacheQueryResultContext, "cached-queries", fmt.Sprintf("%s-%s", sessionID, queryID), int64(len(resultMsgBytes)), bytes.NewReader(resultMsgBytes))
	}
}

func (s *Service) sendStatus(status string, sessionID *string, queryID *string) {
	msg := entity.StatusMessageStruct{
		Type: "query_status_update",
		Value: entity.Status{
			Status:  status,
			QueryID: *queryID,
		},
	}

	msgBytes, err := json.Marshal(msg)
	if err != nil {
		errorhandler.FailWithError(err, "Marshalling status update message went wrong")
		return
	}

	s.producer.PublishMessage(&msgBytes, sessionID)
}
